//
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#ifndef _SEARCH_HXX
#define _SEARCH_HXX

// the definition of the node types can be found for example in
// "Multi-Cut alpha/beta-pruning in game-tre search"

typedef enum _nodeType {
    nodePV,
    nodeCUT,
    nodeALL
} nodeType;


// Killer moves, 2 per depth

class _killers {
public:
    void raz(void) {
        Killer1.m = 0;
        Killer2.m = 0;
    }
    void INLINE set(const CM *pcm) {
        if( !(Killer1.m == pcm->m ) ) {
            Killer2 = Killer1;
            Killer1 = *pcm;
        }
    }
    CM Killer1;
    CM Killer2;
};
extern _killers Killers[];
extern _killers MateKillers[];

// it is not clear atm which history indexing is better

//#define HIST_SIZE       (64*64)
//#define HIST_INDEX(m)   (From(m)+64*To(m))
#define HIST_SIZE       (2*16*64)
#define HIST_INDEX(m)   ((Pc(m)+16*To(m))*2 + (Capt(m) ? 1 : 0))

class _History {
private:
    unsigned int good_hit[HIST_SIZE];
//    unsigned int depth_hit[HIST_SIZE];
    int move_count[HIST_SIZE];
    int failed_high[HIST_SIZE];
    int failed_low_count[HIST_SIZE];
    int shift_count;
public:
    _History() {};
    void raz(void) {
        shift_count = 0;
        for(int i = 0; i < HIST_SIZE; i++) {
            good_hit[i] = move_count[i] = failed_high[i] = failed_low_count[i] = 0;
        }
    }
    // increment the history counter on failed high
    void INLINE update(const CM *cm, int depth) {
        int idx = HIST_INDEX(cm->m);
        good_hit[ idx ] += depth*depth;
        if( good_hit[ idx ] >= 0x00100000 ) {
            shift_count++;
            for(int i = 0; i < HIST_SIZE ; i++ ) {
                good_hit[ i ] = (good_hit[ i ] + 1) / 2;
            }
        }
    }
    void seen(const CM *cm, nodeType nt) {
        if( cm != NULL /*&& nt != nodeALL*/ ) {
            int idx = HIST_INDEX(cm->m);
            move_count[ idx ]++;
        }
    }
    int seen_count(const CM *cm) const {
        return move_count[ HIST_INDEX(cm->m) ];
    }
    void failed_low(const CM *cm, nodeType nt) {
        if( cm != NULL && nt != nodeALL ) {
            int idx = HIST_INDEX(cm->m);
            failed_low_count[ idx ]++;
        }
    }
    int get_ratio(const CM *cm) const {
        int idx = HIST_INDEX(cm->m);
        if( move_count [ idx ] == 0 )
            return 1024;
        return (1024 * failed_high[ idx ]) / move_count [ idx ];
//        return good_count[ idx ] - bad_count [ idx ];
    }
    unsigned int sort_key(const CM *cm) const {
        int idx = HIST_INDEX(cm->m);
        return good_hit[ idx ];
    }
    unsigned int get_stats(const CM *cm) const {
        int idx = HIST_INDEX(cm->m);
        return (failed_high[ idx ] << 16) | failed_low_count[ idx ];
    }
    bool can_reduce(const CM *cm) const {
        if( (cm->cmk & cmkMASK) > cmkNONE )
            return false;
        if( (cm->cmk & cmkMASK) == cmkBADCAPT )
            return false;
        // moves in next category failed high at some point
//        if( (cm->cmk & ~cmkMASK) >= 256 )
//            return false;
        int idx = HIST_INDEX(cm->m);
        if( move_count [ idx ] == 0 )
            return false;
        // we don't want to reduce if there is less than xx percent of time the move has
        // failed low
        // ... fl count is incremented when not @ all nodes, move count is allways incremented...
        // 50,70 => wrong score (fh>1,fl<95) (>3,<85) (3,50)
#if 0
        if( failed_low_count [ idx ] == 0 )
            return false;
        if( ((100*failed_high[ idx ]) / failed_low_count [ idx ]) >= 20)
            return false;
//        if( ((100*failed_high[ idx ]) / move_count [ idx ]) > 10)
//            return false;
//        if( ((100*failed_low_count[ idx ]) / move_count [ idx ]) < 60)
//            return false;
#endif
//        return false;
        // new code : we have failed_high count and move_count
        if( ((1024*failed_high[ idx ]) / move_count [ idx ]) >= 500)
            return false;

        return true;
    }
    bool can_prune(const CM *cm) const {
        if( (cm->cmk & cmkMASK) > cmkNONE )
            return false;
        if( (cm->cmk & cmkMASK) == cmkBADCAPT )
            return false;
        // moves in next category failed high at some point
//        if( (cm->cmk & ~cmkMASK) >= 256 )
//            return false;
        int idx = HIST_INDEX(cm->m);
        if( move_count [ idx ] == 0 )
            return false;
        // we don't want to reduce if there is less than xx percent of time the move has
        // failed low
//        return false;
        // new code : we have failed_high count and move_count
        if( ((1024*failed_high[ idx ]) / move_count [ idx ]) >= 2)
            return false;

        return true;
    }
};

extern _History History;


#endif // _SEARCH_HXX

